@@ -86,6 +86,7 @@ int master_add_autofs_point(struct master_mapent *entry, unsigned logopt,
 		free(ap);
 		return 0;
 	}
+	ap->pref = NULL;
 
 	ap->entry = entry;
 	ap->exp_thread = 0;
@@ -119,6 +120,7 @@ int master_add_autofs_point(struct master_mapent *entry, unsigned logopt,
 	ap->submount = submount;
 	INIT_LIST_HEAD(&ap->mounts);
 	INIT_LIST_HEAD(&ap->submounts);
+	INIT_LIST_HEAD(&ap->amdmounts);
 	ap->shutdown = 0;
 
 	status = pthread_mutex_init(&ap->mounts_mutex, NULL);
@@ -135,15 +137,32 @@ int master_add_autofs_point(struct master_mapent *entry, unsigned logopt,
 
 void master_free_autofs_point(struct autofs_point *ap)
 {
+	struct list_head *p, *head;
 	int status;
 
 	if (!ap)
 		return;
 
+	mounts_mutex_lock(ap);
+	head = &ap->amdmounts;
+	p = head->next;
+	while (p != head) {
+		struct amd_entry *entry = list_entry(p, struct amd_entry, entries);
+		p = p->next;
+		if (!list_empty(&entry->ext_mount))
+			ext_mount_remove(&entry->ext_mount, entry->fs);
+		if (!list_empty(&entry->entries))
+			list_del(&entry->entries);
+		free(entry);
+	}
+	mounts_mutex_unlock(ap);
+
 	status = pthread_mutex_destroy(&ap->mounts_mutex);
 	if (status)
 		fatal(status);
 
+	if (ap->pref)
+		free(ap->pref);
 	free(ap->path);
 	free(ap);
 }
@@ -178,6 +197,8 @@ master_add_map_source(struct master_mapent *entry,
 			return NULL;
 		}
 		source->format = nformat;
+		if (!strcmp(nformat, "amd"))
+			source->flags |= MAP_FLAG_FORMAT_AMD;
 	}
 
 	source->age = age;
@@ -190,6 +211,8 @@ master_add_map_source(struct master_mapent *entry,
 	}
 	source->argc = argc;
 	source->argv = tmpargv;
+	if (source->argv[0])
+		source->name = strdup(source->argv[0]);
 
 	master_source_writelock(entry);
 
@@ -312,6 +335,8 @@ static void __master_free_map_source(struct map_source *source, unsigned int fre
 		free(source->type);
 	if (source->format)
 		free(source->format);
+	if (source->name)
+		free(source->name);
 	if (free_cache && source->mc)
 		cache_release(source);
 	if (source->lookup) {
@@ -430,6 +455,8 @@ master_add_source_instance(struct map_source *source, const char *type, const ch
 			return NULL;
 		}
 		new->format = nformat;
+		if (!strcmp(nformat, "amd"))
+			new->flags |= MAP_FLAG_FORMAT_AMD;
 	}
 
 	new->age = age;
@@ -445,6 +472,8 @@ master_add_source_instance(struct map_source *source, const char *type, const ch
 	}
 	new->argc = argc;
 	new->argv = tmpargv;
+	if (source->name)
+		new->name = strdup(source->name);
 
 	status = pthread_mutex_lock(&instance_mutex);
 	if (status)
@@ -687,6 +716,34 @@ struct autofs_point *master_find_submount(struct autofs_point *ap, const char *p
 	return submount;
 }
 
+struct amd_entry *__master_find_amdmount(struct autofs_point *ap, const char *path)
+{
+	struct list_head *head, *p;
+
+	head = &ap->amdmounts;
+	list_for_each(p, head) {
+		struct amd_entry *entry;
+
+		entry = list_entry(p, struct amd_entry, entries);
+
+		if (!strcmp(entry->path, path))
+			return entry;
+	}
+
+	return NULL;
+}
+
+struct amd_entry *master_find_amdmount(struct autofs_point *ap, const char *path)
+{
+	struct amd_entry *entry;
+
+	mounts_mutex_lock(ap);
+	entry = __master_find_amdmount(ap, path);
+	mounts_mutex_unlock(ap);
+
+	return entry;
+}
+
 struct master_mapent *master_new_mapent(struct master *master, const char *path, time_t age)
 {
 	struct master_mapent *entry;
@@ -865,10 +922,7 @@ int master_read_master(struct master *master, time_t age, int readall)
 		master_mount_mounts(master, age, readall);
 	else {
 		master->read_fail = 0;
-		/* HUP signal sets readall == 1 only */
 		if (!readall)
-			return 0;
-		else
 			master_mount_mounts(master, age, readall);
 	}
 
@@ -1658,7 +1712,7 @@ int master_done(struct master *master)
 	return res;
 }
 
-unsigned int master_get_logopt(void)
+inline unsigned int master_get_logopt(void)
 {
 	return master_list ? master_list->logopt : LOGOPT_NONE;
 }
