@@ -509,8 +509,7 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 				fstype[typelen] = '\0';
 			} else if (_strncmp("nonstrict", cp, 9) == 0) {
 				nonstrict = 1;
-			} else if (_strncmp("strict", cp, 6) == 0 &&
-				   comma - cp == 6) {
+			} else if (_strncmp("strict", cp, 6) == 0) {
 				nonstrict = 0;
 			} else if (_strncmp("nobrowse", cp, 8) == 0 ||
 				   _strncmp("browse", cp, 6) == 0 ||
@@ -757,8 +756,6 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 
 	mc = source->mc;
 
-	memset(m_mapent, 0, MAPENT_MAX_LEN + 1);
-
 	/* Internal hosts map may have loc == NULL */
 	if (!*path) {
 		error(ap->logopt,
@@ -785,7 +782,7 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 	if (*myoptions)
 		m_options_len = strlen(myoptions) + 2;
 
-	m_mapent_len = loc ? strlen(loc) : 0;
+	m_mapent_len = strlen(loc);
 	if (m_mapent_len + m_options_len > MAPENT_MAX_LEN) {
 		error(ap->logopt, MODPREFIX "multi mount mapent too long");
 		return CHE_FAIL;
@@ -796,20 +793,16 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 		strcat(m_mapent, myoptions);
 		if (loc) {
 			strcat(m_mapent, " ");
-			if (loc)
-				strcat(m_mapent, loc);
+			strcat(m_mapent, loc);
 		}
-	} else {
-		if (loc)
-			strcpy(m_mapent, loc);
-	}
+	} else
+		strcpy(m_mapent, loc);
 
 	ret = cache_update_offset(mc, name, m_key, m_mapent, age);
-	if (ret == CHE_DUPLICATE) {
+	if (ret == CHE_DUPLICATE)
 		warn(ap->logopt, MODPREFIX
 		     "syntax error or duplicate offset %s -> %s", path, loc);
-		ret = CHE_OK;
-	} else if (ret == CHE_FAIL)
+	else if (ret == CHE_FAIL)
 		debug(ap->logopt, MODPREFIX
 		      "failed to update multi-mount offset %s -> %s", path, m_mapent);
 	else {
@@ -930,15 +923,9 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 	l = chunklen(p, check_colon(p));
 	loc = dequote(p, l, logopt);
 	if (!loc) {
-		if (strstr(myoptions, "fstype=autofs") &&
-		    strstr(myoptions, "hosts")) {
-			warn(logopt, MODPREFIX "possible missing location");
-			free(myoptions);
-			return 0;
-		}
-		*options = myoptions;
-		*location = NULL;
-		return (p - ent);
+		warn(logopt, MODPREFIX "possible missing location");
+		free(myoptions);
+		return 0;
 	}
 
 	/* Location can't begin with a '/' */
@@ -966,15 +953,10 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		l = chunklen(p, check_colon(p));
 		ent_chunk = dequote(p, l, logopt);
 		if (!ent_chunk) {
-			if (strstr(myoptions, "fstype=autofs") &&
-			    strstr(myoptions, "hosts")) {
-				warn(logopt, MODPREFIX
-				     "null location or out of memory");
-				free(myoptions);
-				free(loc);
-				return 0;
-			}
-			goto next;
+			warn(logopt, MODPREFIX "null location or out of memory");
+			free(myoptions);
+			free(loc);
+			return 0;
 		}
 
 		/* Location can't begin with a '/' */
@@ -1010,7 +992,7 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		strcat(loc, ent_chunk);
 
 		free(ent_chunk);
-next:
+
 		p += l;
 		p = skipspace(p);
 	}
@@ -1111,9 +1093,7 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 				cache_delete_offset_list(me->mc, name);
 				return 1;
 			}
-			ro_len = 0;
-			if (ro_loc)
-				ro_len = strlen(ro_loc);
+			ro_len = strlen(ro_loc);
 
 			tmp = alloca(mnt_root_len + 2);
 			strcpy(tmp, mnt_root);
@@ -1124,8 +1104,7 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 			rv = sun_mount(ap, root, name, namelen, ro_loc, ro_len, myoptions, ctxt);
 
 			free(myoptions);
-			if (ro_loc)
-				free(ro_loc);
+			free(ro_loc);
 		}
 
 		if (ro && rv == 0) {
@@ -1235,12 +1214,12 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 	macro_lock();
 
-	ctxt->subst = addstdenv(ctxt->subst, NULL);
+	ctxt->subst = addstdenv(ctxt->subst);
 
 	mapent_len = expandsunent(mapent, NULL, name, ctxt->subst, slashify);
 	if (mapent_len == 0) {
 		error(ap->logopt, MODPREFIX "failed to expand map entry");
-		ctxt->subst = removestdenv(ctxt->subst, NULL);
+		ctxt->subst = removestdenv(ctxt->subst);
 		macro_unlock();
 		pthread_setcancelstate(cur_state, NULL);
 		return 1;
@@ -1250,7 +1229,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	if (!pmapent) {	
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 		logerr(MODPREFIX "alloca: %s", estr);
-		ctxt->subst = removestdenv(ctxt->subst, NULL);
+		ctxt->subst = removestdenv(ctxt->subst);
 		macro_unlock();
 		pthread_setcancelstate(cur_state, NULL);
 		return 1;
@@ -1258,7 +1237,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	pmapent[mapent_len] = '\0';
 
 	expandsunent(mapent, pmapent, name, ctxt->subst, slashify);
-	ctxt->subst = removestdenv(ctxt->subst, NULL);
+	ctxt->subst = removestdenv(ctxt->subst);
 
 	macro_unlock();
 	pthread_setcancelstate(cur_state, NULL);
@@ -1364,26 +1343,32 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			strcat(m_root, name);
 		}
 
+		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 		cache_readlock(mc);
 		me = cache_lookup_distinct(mc, name);
-		if (me) {
-			/* So we know we're the multi-mount root */
-			if (!me->multi)
-				me->multi = me;
-		}
-
 		if (!me) {
 			free(options);
 			cache_unlock(mc);
+			pthread_setcancelstate(cur_state, NULL);
 			error(ap->logopt,
 			      MODPREFIX "can't find multi root %s", name);
 			return 1;
 		}
 
+		cache_multi_writelock(me);
+		/* Someone beat us to it, return success */
+		if (me->multi) {
+			free(options);
+			cache_multi_unlock(me);
+			cache_unlock(mc);
+			pthread_setcancelstate(cur_state, NULL);
+			return 0;
+		}
+		/* So we know we're the multi-mount root */
+		me->multi = me;
+
 		age = me->age;
 
-		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		cache_multi_writelock(me);
 		/* It's a multi-mount; deal with it */
 		do {
 			char *path, *myoptions, *loc;
@@ -1416,13 +1401,16 @@ int parse_mount(struct autofs_point *ap, const char *name,
 
 			l = parse_mapent(p, options, &myoptions, &loc, ap->logopt);
 			if (!l) {
-				cache_delete_offset_list(mc, name);
-				cache_multi_unlock(me);
-				cache_unlock(mc);
-				free(path);
-				free(options);
-				pthread_setcancelstate(cur_state, NULL);
-				return 1;
+				if (!(strstr(myoptions, "fstype=autofs") &&
+				      strstr(myoptions, "hosts"))) {
+					cache_delete_offset_list(mc, name);
+					cache_multi_unlock(me);
+					cache_unlock(mc);
+					free(path);
+					free(options);
+					pthread_setcancelstate(cur_state, NULL);
+					return 1;
+				}
 			}
 
 			p += l;
@@ -1443,14 +1431,12 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				free(path);
 				free(options);
 				free(myoptions);
-				if (loc)
-					free(loc);
+				free(loc);
 				pthread_setcancelstate(cur_state, NULL);
 				return 1;
 			}
 
-			if (loc)
-				free(loc);
+			free(loc);
 			free(path);
 			free(myoptions);
 		} while (*p == '/' || (*p == '"' && *(p + 1) == '/'));
